import typing as t

import tinycss2
import tinycss2.ast as c2ast


Rule = c2ast.AtRule | c2ast.QualifiedRule
LineNodes = Rule | c2ast.Declaration


def strip_whitespace(content: list[c2ast.Node], beginning: bool = True, end: bool = True):
    """
    Based on the @beginning and @end flags, return a version of @content with
    leading and/or trailing whitespace Nodes removed.
    """
    start = 0
    stop = len(content) - 1
    if beginning:
        while start < stop:
            if not isinstance(content[start], c2ast.WhitespaceToken):
                break
            start += 1
    if end:
        while stop > start:
            if not isinstance(content[stop], c2ast.WhitespaceToken):
                break
            stop -= 1
    return content[start:stop+1]


def mk_whitespace(node: c2ast.Node, whitespace: str):
    """
    Create a whitespace Node with the given @whitespace literal, deriving
    source line and column from @node.
    """
    if isinstance(node, Rule) and node.content:
        line = node.content[-1].source_line
        col = node.content[-1].source_column
    else:
        line = node.source_line
        col = node.source_column

    return c2ast.WhitespaceToken(
        line,
        col,
        whitespace
    )


def wrap_newlines(content: t.Iterable[c2ast.Node]):
    """
    Discard existing whitespace Nodes from @content and ensure there's a newline
    before and after each other Node.
    """
    first = True
    for node in content:
        if first:
            yield mk_whitespace(node, '\n')
            first = False
        if not isinstance(node, c2ast.WhitespaceToken):
            yield node
            if isinstance(node, LineNodes):
                yield mk_whitespace(node, '\n')


def find_selector_start(index: int, content: list[c2ast.Node]):
    """
    Search backwards in @content, starting from @index, and find a starting
    point for a selector block. Works by looking for the end of the prior
    statement or block.
    """
    while index:
        node = content[index]
        if isinstance(node, c2ast.LiteralToken) and node == ';':
            index += 1
            break
        if isinstance(node, c2ast.CurlyBracketsBlock):
            index += 1
            break
        index -= 1
    while isinstance(content[index], c2ast.WhitespaceToken):
        index += 1
    return index


def split_selector(content: list[c2ast.Node]):
    """
    Split a comma-separated selector up into indivual selectors.
    """
    selector: list[c2ast.Node] = []
    for node in content:
        if isinstance(node, c2ast.LiteralToken) and node == ',':
            yield strip_whitespace(selector, end=False)
            selector = []
        else:
            selector.append(node)

    if selector:
        yield strip_whitespace(selector, end=False)


def merge_selectors(sel_1: list[c2ast.Node], sel_2: list[c2ast.Node]):
    """
    Merge two selectors into one. Intended for use with nested selectors. The
    first segment of @sel_2 will not be merged with the last segment of @sel_1
    unless it is a pseudo-class/pseudo-element or a literal "&" begins it.
    For example:
    >>> merge_selectors('body.home #banner', '.wide-logo')
    'body.home #banner .wide-logo'
    >>> merge_selectors(['body.home', '#banner'], ['::after'])
    'body.home #banner::after'
    >>> merge_selectors('body.home #banner', '& .wide-logo')
    'body.home #banner.wide-logo'
    >>> merge_selectors('body.home #banner', '& .wide-logo a')
    'body.home #banner.wide-logo a'
    >>> merge_selectors('body.home #banner', '& a')
    <AssertionError>
    """
    sel_1 = strip_whitespace(sel_1)
    sel_2 = strip_whitespace(sel_2)
    first = sel_2[0]
    if isinstance(first, c2ast.LiteralToken) and first.value in {'&', ':'}:
        if first == '&':
            sel_2 = strip_whitespace(sel_2[1:])
            assert not isinstance(sel_2[0], c2ast.IdentToken)
    else:
        sel_1.append(mk_whitespace(sel_1[-1], ' '))

    sel_2.append(mk_whitespace(sel_2[-1], ' '))
    return [*sel_1, *sel_2]


def pump_at_rule(parent: c2ast.QualifiedRule, at_rule: c2ast.AtRule):
    """
    Propagate an AtRule nested within a another block outwards, so that
    @parent's selectors are attached to all of @at_rule's children and the
    AtRule can live at the top level as CSS3 requires it to.
    """
    new_qualified = c2ast.QualifiedRule(
        at_rule.source_line,
        at_rule.source_column,
        parent.prelude,
        []
    )
    new_at = c2ast.AtRule(
        at_rule.source_line,
        at_rule.source_column,
        at_rule.at_keyword,
        at_rule.lower_at_keyword,
        at_rule.prelude,
        [new_qualified]
    )

    for node in t.cast(list[c2ast.Node], at_rule.content):
        if isinstance(node, c2ast.QualifiedRule):
            new_at.content.extend(
                flatten_qual(
                    merge_selectors(parent.prelude, node.prelude),
                    node.content,
                    parse_declarations=False
                )
            )
        else:
            new_qualified.content.append(node)

    new_qualified.content = list(wrap_newlines(tinycss2.parse_declaration_list(new_qualified.content)))
    new_at.content = list(wrap_newlines(new_at.content))
    return new_at


def flatten_at(rule: c2ast.AtKeywordToken, sel: list[c2ast.Node], body: list[c2ast.Node]):
    """
    Handle de-nesting the children of an AtRule. Only produces one AtRule
    because CSS3 accepts a single layer of nesting within AtRules.
    """
    at_rule = c2ast.AtRule(
        rule.source_line,
        rule.source_column,
        rule.value,
        rule.lower_value,
        list(sel),
        []
    )
    at_rule.content.extend(flatten_children(at_rule, body))
    at_rule.content = list(wrap_newlines(at_rule.content))
    yield at_rule


def flatten_qual(sel: list[c2ast.Node], body: list[c2ast.Node], parse_declarations=True):
    """
    Handle de-nesting the children of a QualifiedRule. Will split nested
    QualifiedRules into their own blocks as well as pumping AtRules (see
    `pump_at_rule()` for an explanation).
    """
    rule = c2ast.QualifiedRule(
        sel[0].source_line,
        sel[0].source_column,
        list(sel),
        []
    )
    children: list[Rule] = []
    for child in flatten_children(rule, body):
        if isinstance(child, c2ast.QualifiedRule):
            child.prelude = merge_selectors(sel, child.prelude)
            children.append(child)
        elif isinstance(child, c2ast.AtRule):
            children.append(pump_at_rule(rule, child))
    rule.prelude = strip_whitespace(rule.prelude, end=False)
    if parse_declarations:
        rule.content = tinycss2.parse_declaration_list(rule.content)
    rule.content = list(wrap_newlines(rule.content))
    yield rule
    yield from children


def flatten_children(parent: Rule, content: list[c2ast.Node]):
    """
    Shared behavior between `flatten_at()` and `flatten_qual()` that adds any
    Nodes that do not need de-nesting to @parent.content, and yields any new
    top-level at-rules/qualified rules.
    """
    last_safe = 0
    for i, node in enumerate(content):
        if isinstance(node, c2ast.CurlyBracketsBlock):
            start = find_selector_start(i-1, content)
            assert start >= last_safe
            parent.content.extend(strip_whitespace(content[last_safe:start]))
            sel = strip_whitespace(content[start:i], end=False)
            if isinstance(sel[0], c2ast.AtKeywordToken):
                yield from flatten_at(sel[0], sel[1:], node.content)
            else:
                for sub_sel in split_selector(sel):
                    yield from flatten_qual(sub_sel, node.content)
            last_safe = i + 1

    parent.content.extend(strip_whitespace(content[last_safe:]))


def flatten_one(node: c2ast.Node):
    """
    Handle de-nesting and parsing for a single node of any type, discarding
    whitespace.
    """
    if isinstance(node, c2ast.AtRule) and node.content:
        token = c2ast.AtKeywordToken(node.source_line, node.source_column, node.at_keyword)
        yield from flatten_at(token, node.prelude, node.content)
    elif isinstance(node, c2ast.QualifiedRule):
        yield from flatten_qual(node.prelude, node.content)
    elif not isinstance(node, c2ast.WhitespaceToken):
        yield node


def flatten_all(nodes: t.Iterable[c2ast.Node]):
    """
    Parse and de-nest a series of Nodes.
    """
    for node in nodes:
        yield from wrap_newlines(flatten_one(node))


def process(code: str):
    """
    Parse a string of CSS, de-nest it, and return a string of new code.
    """
    return tinycss2.serialize(flatten_all(tinycss2.parse_stylesheet(code)))



import re
import typing as t
from pathlib import Path

from .core import Context, ContextDir


def _trim_ext_prefix(path: Path, match: re.Match[str]):
    _groups = match.groupdict()
    if 'stem' in _groups:
        return path.with_stem(_groups['stem'])
    if 'ext' in _groups and _groups['ext']:
        return path.with_name(path.name[:-len(_groups['ext'])])
    return path


def _to_dir_inner(dest: Path, ext: str | None, context: Context, path: Path, match: t.Any):
    path = _trim_ext_prefix(path, match) if ext and isinstance(match, re.Match) else path

    rel = path.relative_to(
        context['input_dir']
        if path.is_relative_to(context['input_dir'])
        else context['working_dir']
    )
    new_path = dest / rel

    if ext:
        new_path = new_path.with_suffix(ext)

    return new_path


def to_dir(dest: Path, ext: str | None = None):
    """
    Factory for PathCalculators that make their input paths children of @dest.
    If @ext is specified, it will replace the extension of input paths. If the
    matcher produced an re.Match, it will be checked for explicitly defined
    extension information for the input paths, allowing for extensions like
    `.tar.gz`.
    """
    def inner(context: Context, path: Path, match: t.Any):
        return _to_dir_inner(dest, ext, context, path, match)

    return inner


def to_output(ext: str | None = None):
    """
    Factory for PathCalculators that make their input paths children of their
    Context's output dir. If @ext is specified, it will replace the extension
    of input paths. If the matcher produced an re.Match, it will be checked for
    explicitly defined extension information for the input paths, allowing for
    extensions like `.tar.gz`.
    """
    def inner(context: Context, path: Path, match: t.Any):
        return _to_dir_inner(context['output_dir'], ext, context, path, match)

    return inner


def to_working(ext: str | None = None):
    """
    Factory for PathCalculators that make their input paths children of their
    Context's working dir. If @ext is specified, it will replace the extension
    of input paths. If the matcher produced an re.Match, it will be checked for
    explicitly defined extension information for the input paths, allowing for
    extensions like `.tar.gz`.
    """
    def inner(context: Context, path: Path, match: t.Any):
        return _to_dir_inner(context['working_dir'], ext, context, path, match)

    return inner


def match_re(re_string: str, re_flags: int = 0, dir: ContextDir | None = None):
    """
    Path matcher using regular expressions. @re_flags will be passed to
    `re.compile()`. @dir, if specified, should be a key to a configured
    directory, not a Path, and will be used to handle matching the beginning of
    Paths; this can be used to avoid pitfalls with unexpected characters in
    input or working directories.
    """
    regex = re.compile(re_string, re_flags)
    def match_func(context: Context, path: Path):
        if dir:
            # Handle this part of matching outside the regex.
            if not path.is_relative_to(context[dir]):
                return None
            path = path.relative_to(context[dir])
        return regex.match(path.as_posix())
    return match_func



from __future__ import annotations

import shutil
import typing as t
from pathlib import Path

if t.TYPE_CHECKING:
    from .core import Context


def direct_copy_step(context: Context, path: Path, output_paths: list[Path]):
    """
    A simple Step which only copies a file to the output directory without
    renaming or extension changes.
    """
    for target_path in output_paths:
        target_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(path, target_path)



from __future__ import annotations

from pathlib import Path

from ..core import Context
from .parser import process


def anchovy_css_step(context: Context, path: Path, output_paths: list[Path]):
    """
    Simple Step to preprocess an Anchovy CSS file into compliant CSS.
    """
    if not output_paths:
        return
    processed = process(path.read_text())
    for target_path in output_paths:
        target_path.parent.mkdir(parents=True, exist_ok=True)
        target_path.write_text(processed)



from pathlib import Path

from anchovy.core import InputBuildSettings, Rule
from anchovy.helpers import match_re, to_output
from anchovy.jinja import JinjaMarkdownStep
from anchovy.simple import direct_copy_step


# Optional, and can be overridden with CLI arguments.
SETTINGS = InputBuildSettings(
    input_dir=Path(__file__).parent / 'basic_site',
    output_dir=Path('output/basic_site'),
)
RULES = [
    # Ignore dotfiles in both the input_dir and the working dir.
    Rule(match_re(r'(.*/)*\..*', dir='input_dir'), None),
    Rule(match_re(r'(.*/)*\..*', dir='working_dir'), None),
    # Render markdown files, then stop processing them.
    Rule(match_re(r'.*\.md'), [to_output('.html'), None], JinjaMarkdownStep()),
    # Copy everything else in static/ directories through.
    Rule(match_re(r'(.*/)*static/.*', dir='input_dir'), to_output(), direct_copy_step),
]



from __future__ import annotations

import argparse
import contextlib
import importlib
import tempfile
import typing as t
from pathlib import Path

from .core import BuildSettings, Context, InputBuildSettings, Rule


class BuildNamespace:
    """
    Internal used to preserve typing between InputBuildSettings, argparse, and
    BuildSettings.
    """
    input_dir: Path
    output_dir: Path
    working_dir: Path | None
    purge_dirs: bool

    def __init__(self, settings: InputBuildSettings | None = None):
        if settings:
            self.__dict__.update(settings)

    # FIXME: Gross :(
    def to_build_settings(self, working_dir: Path):
        return BuildSettings(
            input_dir=self.input_dir,
            output_dir=self.output_dir,
            working_dir=working_dir,
            purge_dirs=self.purge_dirs
        )


@contextlib.contextmanager
def _wrap_temp(path: Path | None):
    if path:
        yield path
    else:
        with tempfile.TemporaryDirectory() as temp_dir:
            yield Path(temp_dir)


def parse_settings_args(settings: InputBuildSettings | None = None, argv: list[str] | None = None, **kw):
    """
    Internal function used by `run_from_rules()` to combine an instance of
    InputBuildSettings with CLI arguments to produce a BuildNamespace, which
    can be easily turned into BuildSettings.
    """
    namespace = BuildNamespace(settings)

    parser = argparse.ArgumentParser(**kw)
    parser.add_argument('-i', '--input',
                        help='input directory with raw files to process',
                        type=Path,
                        dest='input_dir',
                        default=Path('site'))
    parser.add_argument('-o', '--output',
                        help='output directory for final built files',
                        type=Path,
                        dest='output_dir',
                        default=Path('output'))

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-w', '--working',
                       help='directory for intermediate files; defaults to a new temporary directory',
                       type=Path,
                       dest='working_dir')
    group.add_argument('--use-temporary',
                       help='force use of a temporary directory for intermediate files',
                       action='store_const',
                       dest='working_dir',
                       const=None)

    parser.add_argument('--purge',
                        help='purge the output directory before building',
                        action=argparse.BooleanOptionalAction,
                        dest='purge_dirs',
                        default=True)

    return parser.parse_args(argv, namespace=namespace)


def run_from_rules(settings: InputBuildSettings | None, rules: list[Rule], **kw):
    """
    Build a new Context from Settings, Rules, and command line arguments. Then,
    execute a build using the new Context.
    """
    final_settings = parse_settings_args(settings, **kw)
    with _wrap_temp(final_settings.working_dir) as working_dir:
        context = Context(final_settings.to_build_settings(working_dir), rules)
        context.run()


def main():
    """
    Anchovy main function. Finds or creates a Context using an Anchovy config
    file and command line arguments, then executes a build using it.
    """
    parser = argparse.ArgumentParser(description='Build an anchovy project.')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-m',
                       help='import path of a config file to build',
                       type=importlib.import_module,
                       dest='module',
                       default=None)
    group.add_argument('config_file',
                       nargs='?',
                       help='file path to a config file to build',
                       type=Path,
                       default=None)

    args, remaining = parser.parse_known_args()

    if args.config_file:
        label: str = str(args.config_file)

        namespace: dict[str, t.Any] = {'__file__': label}
        with open(args.config_file) as file:
            # We're basically emulating the python command line here.
            # pylint: disable=exec-used
            exec(file.read(), namespace)

        settings: InputBuildSettings | None = namespace.get('SETTINGS')
        rules: list[Rule] | None = namespace.get('RULES')
        context: Context | None = namespace.get('CONTEXT')
    else:
        label: str = f'-m {args.module.__name__}'
        settings: InputBuildSettings | None = getattr(args.module, 'SETTINGS')
        rules: list[Rule] | None = getattr(args.module, 'RULES')
        context: Context | None = getattr(args.module, 'CONTEXT')

    if context:
        context.run()
    else:
        if not rules:
            raise RuntimeError('Anchovy config files must have a RULES or CONTEXT attribute!')
        run_from_rules(settings, rules, argv=remaining, prog=f'anchovy {label}')



import tinycss2
import tinycss2.ast as c2ast

from anchovy.css.parser import flatten_all, Rule


TEST_STYLES = '''
@import foo;

.orange {
    color: #abc;
}

p {
    color: #fff;
    a {
        color: #55f;
    }
    .red {
        color: #f55;

        background-color: #f55;
    }
    :hover {
        color: #eee;
    }
    .orange, span, & .yellow {
        color: #ff5;
    }
    @media (max-width: 1000px) {
        width: 10px;
        height: 10px;
        span {
            margin: 0;
        }
    }
}

a {
    color: #fff;
    @swizzle .orange;
}

@media (min-width: 800px) and (max-width: 1000px) {
    h1 {
        font-size: 15px;
    }
    h2 {
        font-size: 12px;
        span {
            color: orange;
        }
    }
}

@when media(width >= 400px) and media(pointer: fine) {
    h1 {
        font-size: 20px;
    }
}
@else media(width < 200px) {
    h1 {
        font-style: bold;
    }
}
'''

def display_node(node: c2ast.Node, indent=0):
    if isinstance(node, Rule):
        print('\t'*indent, node, node.prelude)
    elif isinstance(node, c2ast.Declaration):
        print('\t'*indent, node, node.value)
    elif not isinstance(node, c2ast.WhitespaceToken):
        print('\t'*indent, node)
    if isinstance(node, Rule) and node.content:
        for child in node.content:
            display_node(child, indent+1)


if __name__ == '__main__':
    nodes = flatten_all(tinycss2.parse_stylesheet(TEST_STYLES))
    for n in nodes:
        display_node(n)



from __future__ import annotations

import shutil
import typing as t
from pathlib import Path

import commonmark
import commonmark.render.renderer
from jinja2 import Environment, FileSystemLoader, select_autoescape

from .core import Context, Step


class JinjaRenderStep(Step):
    """
    Abstract base class for Steps using Jinja rendering.
    """
    env: Environment

    def __init__(self, env: Environment | None):
        self._temporary_env = env

    def bind(self, context: Context):
        """
        Bind this Step to a specific context. Also initializes a Jinja
        environment if none is set up already.
        """
        super().bind(context)

        if self._temporary_env:
            self.env = self._temporary_env
        else:
            self.env = Environment(
                loader=FileSystemLoader(context['input_dir']),
                autoescape=select_autoescape()
            )

    def render_template(self, template_name: str, meta: dict[str, t.Any], output_paths: list[Path]):
        """
        Look up a Jinja template by name and render it with @meta as
        parameters, then save the result to each of the provided @output_paths.
        """
        if not output_paths:
            return

        template = self.env.get_template(template_name)
        for path in output_paths:
            path.parent.mkdir(parents=True, exist_ok=True)
        template.stream(**meta).dump(str(output_paths[0]), encoding='utf-8')
        for path in output_paths[1:]:
            shutil.copy(output_paths[0], path)


class JinjaMarkdownStep(JinjaRenderStep):
    """
    A Step for rendering Markdown using Jinja templates. Parses according to
    CommonMark and Renders to HTML by default.
    """
    encoding = 'utf-8'

    def __init__(self,
                 default_template: str | None = None,
                 md_parser: commonmark.Parser | None = None,
                 md_renderer: commonmark.render.renderer.Renderer | None = None,
                 jinja_env: Environment | None = None):
        super().__init__(jinja_env)
        self.default_template = default_template
        self.md_parser = md_parser or commonmark.Parser()
        self.md_renderer = md_renderer or commonmark.HtmlRenderer()

    def __call__(self, path: Path, output_paths: list[Path]):
        meta, content = self.extract_metadata(path.read_text(self.encoding))
        ast = self.md_parser.parse(content.strip())
        meta |= {'rendered_markdown': self.md_renderer.render(ast).strip()}

        self.render_template(
            meta.get('template', self.default_template),
            meta,
            output_paths
        )

    def extract_metadata(self, text: str):
        """
        Read metadata from the front of a markdown-formatted text.
        """
        meta = {}
        lines = text.splitlines()

        i = 0
        for line in lines:
            if ':' not in line:
                break
            key, value = line.split(':', 1)
            if not key.isidentifier():
                break

            meta[key.strip()] = value.strip()
            i += 1

        return meta, '\n'.join(lines[i:])



from __future__ import annotations

import abc
import functools
import shutil
import typing as t
from pathlib import Path

try:
    import rich.progress as _rich_progress
except ImportError:
    _rich_progress = None
try:
    import tqdm as _tqdm
except ImportError:
    _tqdm = None


T = t.TypeVar('T')
StepFunc = t.Callable[['Context', Path, list[Path]], t.Any]
UnboundStep = t.Union[StepFunc, 'Step'] | None
PathCalc = t.Callable[['Context', Path, T], Path] | None
BoundStep = t.Callable[[Path, list[Path]], t.Any]
BuildSettingsKey = t.Literal['input_dir', 'output_dir', 'working_dir', 'purge_dirs']
ContextDir = t.Literal['input_dir', 'output_dir', 'working_dir']


class InputBuildSettings(t.TypedDict, total=False):
    """
    TypedDict for defining build settings in an Anchovy config file.
    """
    input_dir: Path
    output_dir: Path
    working_dir: Path | None
    purge_dirs: bool

class BuildSettings(t.TypedDict):
    """
    TypedDict for processed build settings ready for passing to Context.
    """
    input_dir: Path
    output_dir: Path
    working_dir: Path
    purge_dirs: bool


def _progress(iterable: t.Iterable[T], desc: str) -> t.Iterable[T]:
    if _rich_progress:
        yield from _rich_progress.track(iterable, desc)
    elif _tqdm is not None:
        yield from _tqdm.tqdm(iterable, desc)
    else:
        print(desc)
        yield from iterable


def _rm_children(path: Path):
    if not path.exists():
        return
    for child in path.iterdir():
        if child.is_dir():
            shutil.rmtree(child)
        else:
            child.unlink()


class Context:
    """
    A context and configuration class for building Anchovy projects.
    """
    def __init__(self,
                 settings: BuildSettings,
                 rules: list[Rule]):
        self.settings = settings
        self.rules = [(r, self.bind(r.step)) for r in rules]

    @t.overload
    def __getitem__(self, key: ContextDir) -> Path: ...
    @t.overload
    def __getitem__(self, key: t.Literal['purge_dirs']) -> bool: ...
    def __getitem__(self, key):
        return self.settings[key]

    def bind(self, step: UnboundStep) -> BoundStep | None:
        """
        Bind a Step to this Context. Calls the `bind()` method on class Steps,
        and creates a partial for function Steps.
        """
        if not step:
            return None

        if isinstance(step, Step):
            step.bind(self)
            return step

        @functools.wraps(step)
        def bound(path: Path, output_paths: list[Path]):
            return step(self, path, output_paths)
        return bound

    def find_inputs(self, path: Path):
        """
        Overridable function to get paths to process based on a given @path.
        Default behavior is to recursively search for files but exclude the
        directories themselves.
        """
        for candidate in path.iterdir():
            if candidate.is_dir():
                yield from self.find_inputs(candidate)
            else:
                yield candidate

    def process(self, input_paths: list[Path] | None = None):
        """
        Process a set of files using the Context's defined rules. If
        @input_paths is empty or None, `self.find_inputs()` will be used to get
        a tree of files to process. If intermediate files are produced,
        `self.process()` will be called recursively with them.
        """
        input_paths = input_paths or list(self.find_inputs(self.settings['input_dir']))
        # We want to handle tasks in the order they're defined!
        tasks: dict[BoundStep, list[tuple[Path, list[Path]]]]
        tasks = {step: [] for rule, step in self.rules if step}

        for path in _progress(input_paths, 'Planning...'):
            for rule, step in self.rules:
                if match := rule.match(self, path):
                    # None can be used to halt further rule processing.
                    if not step:
                        break
                    output_paths: list[Path] = []
                    for pathcalc in rule.pathcalcs:
                        # None can be used to halt further rule processing in
                        # paths as well. This allows a single rule to both do
                        # processing and also halt further processing.
                        if not pathcalc:
                            break
                        output_paths.append(pathcalc(self, path, match))
                    else:
                        tasks[step].append((path, output_paths))
                        # We didn't break above, avoid the break below!
                        continue
                    tasks[step].append((path, output_paths))
                    # We need two breaks because we're trying to get out of the
                    # surrounding for loop.
                    break

        flattened: list[tuple[BoundStep, Path, list[Path]]] = []
        for step, paths in tasks.items():
            flattened.extend((step, p, ops) for p, ops in paths)

        further_processing: list[Path] = []
        for step, path, output_paths in _progress(flattened, 'Processing...'):
            print(f'{path} ⇒ {", ".join(str(p) for p in output_paths)}')
            step(path, output_paths)
            further_processing.extend(
                p for p in output_paths
                if p.is_relative_to(self['working_dir'])
            )

        if further_processing:
            self.process(further_processing)

    def run(self, input_paths: list[Path] | None = None):
        """
        Execute pre-run hooks (currently only the baked-in directory purge),
        then call `self.process()` with @input_paths.
        TODO: Support custom pre/post hooks.
        """
        if self['purge_dirs']:
            _rm_children(self['output_dir'])
            _rm_children(self['working_dir'])
        self.process(input_paths)


class Rule(t.Generic[T]):
    """
    A single rule for Anchovy file processing, with a matcher, output path
    calculators, and an optional Step to run.
    """
    def __init__(self,
                 match: t.Callable[[Context, Path], T | None],
                 pathcalc: t.Sequence[PathCalc[T] | Path ] | PathCalc[T] | Path,
                 step: UnboundStep = None):
        self.match = match
        self.step = step
        if not isinstance(pathcalc, t.Sequence):
            pathcalc = [pathcalc]
        self.pathcalcs = [self._path_to_pathcalc(p) if isinstance(p, Path) else p for p in pathcalc]

    def _path_to_pathcalc(self, path: Path):
        # pylint: disable=unused-argument
        def wrapper(context: Context, input_path: Path, match: t.Any):
            return path
        return wrapper


class Step(abc.ABC):
    """
    Abstract base class for Steps, individual processing stages used to build a
    full Anchovy ruleset.
    """
    context: Context

    def bind(self, context: Context):
        """
        Bind this Step to a Context.
        """
        self.context = context

    @abc.abstractmethod
    def __call__(self, path: Path, output_paths: list[Path]):
        ...
